/* $Id: install.c,v 1.18 2001/04/18 00:11:24 richdawe Exp $ */

/*
 *  install.c - Install routines for zippo
 *  Copyright (C) 1999-2001 by Richard Dawe
 *      
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

/* libzippo includes */
#include <libzippo/package.h>
#include <libzippo/packlist.h>
#include <libzippo/packdep.h>
#include <libzippo/dsm.h>
#include <libzippo/mft.h>
#include <libzippo/util.h>
#include <libzippo/archive.h>
#include <libzippo/backup.h>

#include "common.h"
#include "zippo.h"
#include "install.h"
#include "md5hash.h"

#ifdef HAVE_CONIO_H
#include <conio.h>
#else /* !HAVE_CONIO_H */
/* This hack needs testing! How do you flush input stream before read? */
#define getch getchar
#define putch putchar
#endif /* HAVE_CONIO_H */

/* -------------------
 * - perform_install -
 * ------------------- */

int
perform_install (ZIPPO_INSTALL *req)
{
  PACKAGE_INFO   *packages = NULL;
  PACKAGE_INFO   *package  = NULL;
  DSM_FILE_ERROR *dfe      = NULL; /* List of err's in main DSM parse. */
  DSM_FILE_ERROR *ndfe     = NULL;
  DSM_ERROR      *de       = NULL; /* List of err's in new DSM parse.  */
  DSM_ERROR      *nde      = NULL;
  PACKAGE_INFO    newpackage;
  char db_path[PATH_MAX];
  char *dsm = NULL;                /* DSM read into memory      */
  int dup_action = DUP_QUERY;  
  int keypress = 0;
  int ret = 1;
  int i;

  /* For dependency checking need a list of failed deps to report to user. */
#define MAX_FAILED_DEP MAX_PACKAGES
  PACKAGE_DEP *failed_dep[MAX_FAILED_DEP + 1]; /* Leave space for NULL */
  int          failed_dep_max = MAX_FAILED_DEP;  

  /* Get a list of installed packages */
  packages = dsm_load_all((const char **) req->dsm_path, NULL, &dfe);
  packages = ver_load_all((const char **) req->mft_path, packages);
	
  packlist_dedupe(packages);
  packlist_xref(packages);
	
  if (packages == NULL)
    warn("No installed packages found");

  /* Display DSM errors. */
  if (dfe != NULL)
    warn("Parsing failed for some DSM(s) in installed database");
	
  for (ndfe = dfe; ndfe != NULL; ndfe = ndfe->q_forw) {
    dsm_perror(ndfe->name, ndfe->de);
  }
	
  dsm_free_file_error_list(dfe);
  dfe = NULL;

  /* Build the DSM database path */
  strcpy(db_path, req->root);
  addforwardslash(db_path);
  strcat(db_path, ZIPPO_DB_PREFIX);
  addforwardslash(db_path);
	
  /* TODO: Handle install from the available list without having to put
   * the whole db path in, e.g. /djgpp/share/zippo/db-avail/whatever.dsm */

  if (dsm_get_and_parse(req->name, &newpackage, &dsm, &de) != DSM_OK) {
    /* If there were any parsing errors, display them. */
    for (nde = de; nde != NULL; nde = (DSM_ERROR *) nde->q_forw) {
      dsm_perror(req->name, nde);
    }

    /* Tidy up */
    dsm_free_error_list(de);
    de = NULL;

    /* Can't continue without a valid DSM. */
    die("Unable to parse DSM!");
  }

  /* TODO: Allow command-line override. */
  dup_action = newpackage.duplicate_action;

  if (isarchive(req->name))
    strcpy(newpackage.path, req->name);

  /* Currently we only support installing binary, source, documentation
   * and virtual packages. */
  switch(newpackage.version.type) {
  case TYPE_BINARIES:
  case TYPE_SOURCES:
  case TYPE_DOCUMENTATION:
  case TYPE_VIRTUAL:
    break;
		
  default:
    dief("Package type '%s' not supported yet for installs!",
	 package_type_string(&newpackage));
    break;
  }
	
  /* Check all archives are available */
  if (   !install_check_archives(&newpackage, req->name,
				 (const char **) req->zip_path)
      && (req->mod != IM_TEST)) {
    /* Tidy up */
    free(dsm);
    packlist_free(packages);
    
    /* Bail */
    return(EXIT_FAILURE);
  }
		
  /* - Check all the dependencies - */

  /* Build xrefs between the new package and the package list, to check
   * that its dependencies will be satisfied. */
  package_xref(packages, &newpackage);

  /* ret == 1 signifies that the checks are successful; 0 indicates an
   * error. */
  ret = 1;

  /* Initialise failed_dep array */
  for (i = 0; i <= failed_dep_max; i++) { failed_dep[i] = NULL; }

  /* Requires */
  if (!package_check_deps(&newpackage, PACKAGE_DEP_REQUIRES,
			  failed_dep, failed_dep_max)) {
    printf("- Required packages not found:\n");
    for (i = 0; failed_dep[i] != NULL; i++) {
      printf("%s\n", package_dep_string(failed_dep[i]));
    }
    ret = 0;
  }

  /* Depends on */
  if (!package_check_deps(&newpackage, PACKAGE_DEP_DEPENDS_ON,
			  failed_dep, failed_dep_max)) {
    printf("- Depends on packages not found - "
	   "functionality will be impaired (continuing):\n");
    for (i = 0; failed_dep[i] != NULL; i++) {
      printf("%s\n", package_dep_string(failed_dep[i]));
    }
  }

  /* Conflicts with */
  if (!package_check_deps(&newpackage, PACKAGE_DEP_CONFLICTS_WITH,
			  failed_dep, failed_dep_max)) {
    printf("- Conflicts with packages found:\n");
    for (i = 0; failed_dep[i] != NULL; i++) {
      printf("%s\n", package_dep_string(failed_dep[i]));
    }
    ret = 0;
  }

  /* Replaces */
  if (!package_check_deps(&newpackage, PACKAGE_DEP_REPLACES,
			  failed_dep, failed_dep_max)) {
    printf("- Replaces packages found (use --upgrade instead):\n");
    for (i = 0; failed_dep[i] != NULL; i++) {
      printf("%s\n", package_dep_string(failed_dep[i]));
    }
    ret = 0;
  }

  /* Install before */
  if (!package_check_deps(&newpackage, PACKAGE_DEP_INSTALL_BEFORE,
			  failed_dep, failed_dep_max)) {
    printf("- Install before packages found:\n");
    for (i = 0; failed_dep[i] != NULL; i++) {
      printf("%s\n", package_dep_string(failed_dep[i]));
    }
    ret = 0;
  }

  /* Install after */
  if (!package_check_deps(&newpackage,
			  PACKAGE_DEP_INSTALL_AFTER,
			  failed_dep, failed_dep_max)) {
    printf("- Install after packages found (no error, continuing):\n");
    for (i = 0; failed_dep[i] != NULL; i++) {
      printf("%s\n", package_dep_string(failed_dep[i]));
    }
  }

  /* Does the package already exist in the database? */
  if (ret == 1) {
    for (package = packages;
	 package != NULL;
	 package = (PACKAGE_INFO *) package->q_forw) {
      if (strcmp(package->name, newpackage.name) != 0)
	continue;      

      if (package_vercmp(&package->version, &newpackage.version) != 0)
	continue;

      printf("Package %s %s (%s) is already installed!\n",
	     newpackage.name,
	     package_version_string(&newpackage.version),
	     package_type_string(&newpackage));
      ret = 0;
    }
  }

  /* Install can't occur because of failed dependencies. */
  if (ret != 1) {
    /* Free memory */
    free(dsm);
    packlist_free(packages);
    return(EXIT_FAILURE);
  }

  /* Show the install warning - confirmation required (see below). */
  if (newpackage.install_warning != NULL)
    printf("*** WARNING ***\n%s\n", newpackage.install_warning);

  /* Are we in test mode? If so, bail out now. */
  if (req->mod == IM_TEST) {    
    /* Tidy up */    
    free(dsm);
    packlist_free(packages);

    info("Install in test-mode complete - exiting\n");
    return(EXIT_SUCCESS);
  }  

  /* Get confirmation, if there was an install warning. */
  if (newpackage.install_warning != NULL) {
    printf("[c]ontinue or [a]bort? ");

    keypress = 0;
    while (keypress == 0) {
      keypress = getch();

      switch(keypress) {
      case 'c': case 'C': case 'a': case 'A':
	break;

      default:
	putch('\a');
	keypress = 0;
	break;
      }
    }

    printf("%c\n", keypress);

    if ((keypress == 'a') || (keypress == 'A')) {
      printf("Installation of %s %s (%s) aborted by user\n",
	     newpackage.name,
	     package_version_string(&newpackage.version),
	     package_type_string(&newpackage));

      /* Free memory */
      free(dsm);
      packlist_free(packages);

      return(EXIT_FAILURE);
    }
  }

  /* - Install - */

  if (!install_extract_archive(&newpackage, req->verbosity, dup_action,
			       req->name, req->prefix, req->backup_prefix,
			       (const char **) req->zip_path)) {
    printf("Installation of %s %s (%s) failed\n",
	     newpackage.name,
	     package_version_string(&newpackage.version),
	     package_type_string(&newpackage));

    /* Tidy up */
    free(dsm);
    packlist_free(packages);

    /* Bail */
    return(EXIT_FAILURE);
  }

  /* Write DSM into db directory. */
  if (!install_dsm(&newpackage, db_path, dsm)) {
    free(dsm);
    packlist_free(packages);
    /* TODO: Better error msg */
    dief("Unable to write DSM for %s %s (%s)!",
	 newpackage.name,
	 package_version_string(&newpackage.version),
	 package_type_string(&newpackage));
  } else {
    printf(". Created DSM for %s %s (%s)\n",
	   newpackage.name,
	   package_version_string(&newpackage.version),
	   package_type_string(&newpackage));
  }
	
  /* Generate MD5 hashes for the installed files. */
  if (md5hash(&newpackage, db_path, req->prefix,
	      (const char **) req->mft_path) == -1) {
    warnf("Error generating MD5 hashes for %s %s (%s)",
	  newpackage.name,
	  package_version_string(&newpackage.version),
	  package_type_string(&newpackage));
    perror("install");
  } else {
    printf(". Created MD5 hashes for files from %s %s (%s)\n",
	   newpackage.name,
	   package_version_string(&newpackage.version),
	   package_type_string(&newpackage));
  }

  /* Create entries in the info directory for any info files. */
  if (!install_info_entries(&newpackage, db_path, req->prefix)) {
    warnf("Error creating info directory entries for %s %s (%s)",
	  newpackage.name,
	  package_version_string(&newpackage.version),
	  package_type_string(&newpackage));
  }
	
  /* Done! */

  /* Free memory */
  free(dsm);
  packlist_free(packages);

  ret = (ret == 1) ? EXIT_SUCCESS : EXIT_FAILURE;
  return(ret);
}

/* --------------------------
 * - install_check_archives -
 * -------------------------- */

int
install_check_archives (PACKAGE_INFO *package,
			const char *req_name,
			const char **archive_paths)
{
  char *a_name = NULL;
  int   ret    = 1; /* Succeed by default */
  int   i;

  /* TODO: tar-gzip, etc. */
		
  /* No zips => no install! */
  if ( (package->zip == NULL) || (package->zip[0] == NULL) ) {
    /* NB: Virtual packages are installed by copying the DSM. */
    if (   (package->version.type == TYPE_BINARIES)
	&& (package->version.type == TYPE_SOURCES)
        && (package->version.type == TYPE_DOCUMENTATION) ) {
      warn("No archives specified in package DSM!");
      ret = 0;
    }
  }

  for (i = 0; package->zip[i] != NULL; i++) {
    /* Barf here if it isn't an archive. */
    if (!isarchive(package->zip[i])) {      
      warnf("'%s' is not an archive", package->zip[i]); 
      ret = 0;
    }
		
    /* Find the archive */
    /* TODO: This should handle the case where we are installing from
     * a renamed archive; if there's only one archive, ask the user if
     * the archive has been renamed. */
    a_name = find_archive(package->zip[i], req_name,
			  (const char **) archive_paths);

    if (a_name == NULL) {
      warnf("Unable to locate archive '%s'!\n", package->zip[i]);
      ret = 0;
    }
  }

  return(ret);
}

/* ---------------------------
 * - install_extract_archive -
 * --------------------------- */

static int
free_toc (char **toc)
{
  int i;

  for (i = 0; toc[i] != NULL; i++) { free(toc[i]); }
  free(toc);
  return(1);
}

int
install_extract_archive (PACKAGE_INFO *package,
			 const int verbosity,
			 const int dup_action,
			 const char *req_name,
			 const char *prefix,
			 const char *backup_prefix,
			 const char **zip_path)
{
  char          filename[PATH_MAX];  /* Temporary-use file name   */
  char          filename2[PATH_MAX]; /* Temporary-use file name 2 */  
  char         *a_name      = NULL;
  char        **a_toc       = NULL;
  const char   *backup_path = NULL;
  int           action      = DUP_QUERY;
  int           extract     = 0;
  int           keypress    = 0;
  mode_t        mode        = 0;
  struct stat   s;
  int           ret         = 0;
  char         *p           = NULL;
  int           i           = 0;
  int           j           = 0;

  /* Generate name for the backup path */
  backup_path = backup_get_path(package, backup_prefix);

  /* Make sure the prefix exists */
  printf(". Creating prefix directory '%s'...", prefix);

  mode = S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH;
  ret = recursive_mkdir(prefix, mode);

  if ((ret != 0) && (errno != EEXIST)) {
    printf("failed\n");
    perror("zippo");
    return(0);
  }

  printf("done\n");

  /* TODO: tar-gzip, etc. */

  for (i = 0; package->zip[i] != NULL; i++) {
    /* Barf here if it isn't an archive. */
    if (!isarchive(package->zip[i])) {
      warnf("'%s' is not an archive", package->zip[i]);
      return(0);
    }
		
    /* Find the archive */
    a_name = find_archive(package->zip[i], req_name, zip_path);

    if (a_name == NULL) {
      printf("Unable to locate archive '%s'!\n", package->zip[i]);
      /* TODO: Make all actions logged => can undo */      
      
      return(0);
    }

    printf(". Installing from archive '%s'\n", a_name);

    /* Get the table of contents */
    a_toc = archive_get_toc(a_name);
    if (a_toc == NULL) {
      warnf("Archive '%s' has no entries\n", a_name);
      continue;
    }

    /* Extract all files */
    for (j = 0; a_toc[j] != NULL; j++) {
      strcpy(filename, prefix);
      strcat(filename, a_toc[j]);

      /* Directory? */
      if (ends_with_slash(filename)) {
	/* Create the directory */
	ret = recursive_mkdir(filename, mode);

	if (ret != 0) {
	  /* Unable to create directory */
	  warnf("Unable to create directory '%s'", filename);	  

	  /* Bail out */
	  free_toc(a_toc), a_toc = NULL;
	  return(0);
	}
				
	if (verbosity != V_NORMAL)
	  printf(". Created directory '%s'\n", filename);
      }

      /* Check the directory stem and create if it doesn't exist. Some zip
       * files seem to contain the directories after the files. Some don't
       * even seem to have directory names. Extracting the file will fail
       * if the directory doesn't exist. */
      strcpy(filename2, filename);
      p = strrchr(filename2, '/');
      if (p != NULL) {
	*p = '\0';
	ret = stat(filename2, &s);

	if ((ret == -1) && (errno = ENOENT)) {
	  ret = recursive_mkdir(filename2, mode);
	  if (ret != 0) {
	    /* Unable to create directory */
	    warnf("Unable to create directory '%s'", filename2);

	    /* Bail out */
	    free_toc(a_toc), a_toc = NULL;
	    return(0);
	  }

	  if (verbosity != V_NORMAL)
	    printf(". Created directory '%s'\n", filename2);
	} else if ((ret == -1) || !S_ISDIR(s.st_mode)) {
	} /* Otherwise we need do nothing. */
      }

      /* If this archive entry is a directory, move onto the next entry. */
      if (strrchr(filename, '/') == filename + strlen(filename) - 1)
	continue;

      /* If the file already exists, behaviour is controlled by
       * the duplicate-action setting. */
      extract = 1;

      if (access(filename, F_OK | R_OK) == 0) {	
	if (dup_action == DUP_QUERY) {
	  /* Prompt user */
	  printf("'%s' already exists\n", filename);
	  printf("[r]eplace, [b]ackup, [s]kip or [a]bort? ");

	  keypress = 0;
	  while (keypress == 0) {
	    keypress = getch();

	    switch(keypress) {
	    case 'r': case 'R': case 'b': case 'B':
	    case 's': case 'S': case 'a': case 'A':
	      break;

	    default:
	      putch('\a');
	      keypress = 0;
	      break;
	    }
	  }

	  printf("%c\n", keypress);

	  switch(keypress) {
	  case 'r': case 'R':
	    action = DUP_REPLACE;
	    break;

	  default:
	  case 'b': case 'B':
	    action = DUP_BACKUP;
	    break;

	  case 's': case 'S':
	    action = DUP_SKIP;
	    break;

	  case 'a': case 'A':
	    action = DUP_ABORT;
	    break;
	  }
	} else {
	  action = dup_action;
	}

	/* Display message for action on duplicate. */
	switch(action) {
	case DUP_REPLACE:
	  printf("'%s' already exists - overwriting\n", filename);
	  break;

	default:
	case DUP_BACKUP:
	  ret = backup_file(a_toc[j], prefix, backup_path);
	  if (ret == 0) {
	    printf(". Backed up: '%s/%s' -> '%s/%s'\n",
		   prefix, a_toc[j], backup_path, a_toc[j]);
	  } else {
	    warnf("Unable to back up '%s/%s' - ABORTING\n", prefix, a_toc[j]);

	    /* Bail out */
	    free_toc(a_toc), a_toc = NULL;
	    return(0);
	  }
	  break;

	case DUP_KEEP:
	case DUP_SKIP:
	  extract = 0;
	  printf("'%s' already exists - skipping\n", filename);
	  break;

	case DUP_ABORT:
	  printf("Installation of %s %s (%s) aborted by user\n",
		 package->name,
		 package_version_string(&package->version),
		 package_type_string(package));

	  /* Bail out */
	  free_toc(a_toc), a_toc = NULL;
	  return(0);
	  break;
	}
      }

      /* Extract file */
      if (extract) {
	if (!archive_extract_file(a_name, a_toc[j], filename)) {
	  /* Unable to extract file! */
	  warnf("Unable to create file '%s' - ABORTING", filename);

	  /* Bail out */
	  free_toc(a_toc), a_toc = NULL;
	  return(0);
	}
			
	if (verbosity != V_NORMAL)
	  printf(". Created file '%s'\n", filename);
      }
    }

    /* Free the TOC */
    free_toc(a_toc), a_toc = NULL;
  }

  return(1);
}

/* ---------------
 * - install_dsm -
 * --------------- */

int
install_dsm (PACKAGE_INFO *package, const char *db_path, const char *dsm)
{
  char  filename[PATH_MAX]; /* Temporary-use file name */
  FILE *fp  = NULL;
  int   ret = 1; /* Succeed by default */

  strcpy(filename, db_path);
  strcat(filename, package->dsm_name);
  if (!isdsm(filename))
    strcat(filename, ".dsm");

  fp = fopen(filename, "wt");
  if (fp == NULL)
    return(0);

  if (fwrite(dsm, strlen(dsm), 1, fp) <= 0) {
    /* Failed */
    ret = 0;
  }

  fclose(fp);
  return(ret);
}

/* ---------------------------------
 * - install_info_entries_internal -
 * --------------------------------- */

/*
 * Create or delete an entry in the info directory file for any info files
 * belonging to the package. This relies on MD5 hashes existing for
 * the package. 'install_flag' should be non-zero if you want to create
 * an entry; conversely it should be zero to delete an entry.
 *
 * On success, 1 is returned. On failure, 0 is returned.
 */

int
install_info_entries_internal (PACKAGE_INFO *package, const char *db_path,
			       const char *prefix, int install_flag)
{
  char             cmd_install_info[PATH_MAX * 4]; /* 3 paths + some space */
  char             info_prefix[PATH_MAX]; /* Path to search for info files */
  const char       info_subdir[] = "info/";
  md5hash_entry  **list          = NULL;
  char            *filename      = NULL;
  char            *p             = NULL;
  int              matched;
  int              ret;
  int              i;

  /* Build info_prefix */
  strcpy(info_prefix, prefix);
  addforwardslash(info_prefix);
  strcat(info_prefix, info_subdir);
  forwardslashify(info_prefix);

  /* Get all the MD5 hash values calculated previously. */
  list = md5hash_get(package, db_path);
  if (list == NULL) {
    /* Failed */
    return(0);
  }

  /* Search for files in info_prefix */
  for (i = 0; list[i] != NULL; i++) {    
    filename = list[i]->filename;
    if (!(strstr(filename, info_prefix) == filename))
      continue;

    /* Find file extension */
    p = strrchr(filename, '.');
    if (p == NULL)
      continue;
    p++;
      
    /* Match info filename extensions */
    matched  = (strcmp(p, "inf") == 0);
    matched |= (strcmp(p, "info") == 0);

    if (!matched)
      continue;

    /* Now run install-info for this info file */
    sprintf(cmd_install_info,
	    "install-info %s--info-file=%s --info-dir=%s",
	    !install_flag ? "--delete " : "", filename, info_prefix);    

    infof("Executing: %s", cmd_install_info);

    ret = system(cmd_install_info);
    if (ret)
      warnf("install-info returned error code %d", ret);
  }

  /* Tidy up */
  md5hash_entries_free(list);
  free(list), list = NULL;

  return(1);
}

/* Wrappers for install_info_entries_internal() */

/* Create */
int
install_info_entries (PACKAGE_INFO *package, const char *db_path,
		      const char *prefix)
{
  return(install_info_entries_internal(package, db_path, prefix, 1));
}

/* Delete */
int
uninstall_info_entries (PACKAGE_INFO *package, const char *db_path,
			const char *prefix)
{
  return(install_info_entries_internal(package, db_path, prefix, 0));
}
