/* $Id: archive.c,v 1.4 2001/04/18 00:04:29 richdawe Exp $ */

/*
 *  archive.c - Archive handling functions for zippo
 *  Copyright (C) 1999-2001 by Richard Dawe
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* libzippo includes */
#include <libzippo/archive.h>
#include <libzippo/util.h>

#include "common.h"
#include "unzip.h"

/* Remember the last opened zip file, so we don't have to keep opening it. */
static unzFile zipfile           = NULL;
static char    zipname[PATH_MAX] = "";
static int     zip_opened        = 0;

/* TODO: tar-gzip */

static int    zip_extract_file  (const char *archive, const char *file, 
                                 const char *dest);
static char **zip_get_toc       (const char *archive);
static int    zip_get_file_info (const char *archive, const char *file,
                                 ARCHIVE_FILE_INFO *afi);

static int    zip_extract_file_to_memory (const char *archive,
					  const char *file,
					  char *buf, unsigned long buflen);

/* TODO: Error codes? */

/* ------------
 * - zip_open -
 * ------------ */

/* This opens the zip archive with "memory" between invocations of the
 * archive_*() functions. */

int __inline__
zip_open (const char *archive)
{
  if (!zip_opened) {
    zipfile = unzOpen(archive);
    if (zipfile == NULL) return(0);
    strcpy(zipname, archive);
    zip_opened = 1;
  } else {
    if (strcmp(archive, zipname) != 0) {
      unzClose(zipfile);
      zip_opened = 0;

      zipfile = unzOpen(archive);
      if (zipfile == NULL) return(0);
      strcpy(zipname, archive);
      zip_opened = 1;
    }
  }

  return(1);
}

/* -------------------
 * - archive_get_toc -
 * ------------------- */

/*
 * Retrieve the table-of-contents (TOC) of the archive, i.e. a list of all
 * files (including directories) inside the archive. This list must be freed
 * by the caller.
 *
 * Returns a NULL-terminated list on success, otherwise NULL on failure.
 */

char **
archive_get_toc (const char *archive)
{
  if (!iszip(archive))
    return(NULL);
  else
    return(zip_get_toc(archive));
}

/* ---------------
 * - zip_get_toc -
 * --------------- */

static char **
zip_get_toc (const char *archive)
{
  char **toc = NULL;
  unz_global_info gi;
  char currentfile[PATH_MAX];
  int ret, i, j;

  if (!zip_open(archive)) return(NULL);

  /* Get the global info */
  if (unzGetGlobalInfo(zipfile, &gi) != UNZ_OK) return(NULL);

  /* Allocate a TOC list */
  toc = (char **) malloc(sizeof(char *) * (gi.number_entry + 1));
  if (toc == NULL) return(NULL);
  for (i = 0; i <= gi.number_entry; i++) { toc[i] = NULL; }

  for (i = 0, ret = unzGoToFirstFile(zipfile);
       ret == UNZ_OK;
       i++, ret = unzGoToNextFile(zipfile)) {
    /* Get the file details */
    ret = unzGetCurrentFileInfo(zipfile, NULL,
				currentfile, sizeof(currentfile),
				NULL, 0, NULL, 0);

    if (ret != UNZ_OK) {
      for (j = 0; toc[j] != NULL; j++) {
	free(toc[j]);
	toc[j] = NULL;
      }
      free(toc);
      return(NULL);
    }

    toc[i] = strdup(currentfile);
  }

  return(toc);
}

/* ------------------------
 * - archive_extract_file -
 * ------------------------ */

/* Extract from the file 'file' from the archive file 'archive' into the
 * destination file 'dest'. This routine figures out which archive format
 * 'archive' is by looking at the file extension. Directories should be
 * created using mkdir(). */

/* TODO: Symlinks? Or not supported? */

int
archive_extract_file (const char *archive, const char *file, const char *dest)
{
  if (!iszip(archive))
    return(0);
  else
    return(zip_extract_file(archive, file, dest));
}

/* --------------------
 * - zip_extract_file -
 * -------------------- */

static int
zip_extract_file (const char *archive, const char *file, const char *dest)
{
  FILE *fp = NULL;
  char  buf[1024];
  int   nread;
  int   ret = 1; /* Succeed by default */

  if (!zip_open(archive))
    return(0);

  /* Find the requested file - 0 = default case sensitivity */
  if (unzLocateFile(zipfile, file, 0) != UNZ_OK)
    return(0);

  /* Now extract */
  fp = fopen(dest, "wb");
  if (fp == NULL)
    return(0);

  if (unzOpenCurrentFile(zipfile) != UNZ_OK) {
    fclose(fp);
    return(0);
  }

  while ( (nread = unzReadCurrentFile(zipfile, buf, sizeof(buf))) > 0) {
    if (fwrite(buf, nread, 1, fp) <= 0) {
      ret = 0;
      break;
    }
  }
	
  unzCloseCurrentFile(zipfile);
  fclose(fp);

  return(ret);
}

/* -------------------------
 * - archive_get_file_info -
 * ------------------------- */

int
archive_get_file_info (const char *archive, const char *file,  
		       ARCHIVE_FILE_INFO *afi)
{
  if (!iszip(archive))
    return(0);
  else
    return(zip_get_file_info(archive, file, afi));
}

/* ---------------------
 * - zip_get_file_info -
 * --------------------- */

static
int zip_get_file_info (const char *archive, const char *file,
		       ARCHIVE_FILE_INFO *afi)
{
  unz_file_info fi;
  char currentfile[PATH_MAX];
  int len;

  if (!zip_open(archive)) return(0);

  /* Find the requested file - 0 = default case sensitivity */
  if (unzLocateFile(zipfile, file, 0) != UNZ_OK) return(0);

  if (unzGetCurrentFileInfo(zipfile, &fi,
			    currentfile, sizeof(currentfile),
			    NULL, 0, NULL, 0) != UNZ_OK)
    return(0);

  /* Store the details */
  len = strlen(currentfile) + 1;
  if (sizeof(afi->name) < len) len = sizeof(afi->name);
  strncpy(afi->name, currentfile, sizeof(afi->name));
  afi->name[len - 1] = '\0';
  afi->size = fi.uncompressed_size;

  return(1);
}

/* ----------------------------------
 * - archive_extract_file_to_memory -
 * ---------------------------------- */

/* This function extracts the given file into a buffer created for it. The
 * caller should free this buffer. */

static char *
__extract_to_memory (const char *archive, const char *file, const int textmode)
{
  ARCHIVE_FILE_INFO afi;
  char          *buf    = NULL;
  unsigned long  buflen = 0;

  /* Create a big enough buffer. */
  if (!archive_get_file_info(archive, file, &afi)) return(NULL);

  buflen = afi.size;
  if (!textmode)
    buf = (char *) malloc(buflen);
  else
    buf = (char *) malloc(buflen + 1); /* Space for nul */

  if (buf == NULL) return(NULL);

  /* TODO: Other formats */
  if (!zip_extract_file_to_memory(archive, file, buf, buflen)) {
    free(buf);
    return(NULL);
  }

  if (textmode) buf[buflen] = '\0';

  return(buf);
}

char *
archive_extract_file_to_memory (const char *archive, const char *file)
{
  return(__extract_to_memory(archive, file, 0));
}

char *
archive_extract_text_file_to_memory (const char *archive, const char *file)
{
  return(__extract_to_memory(archive, file, 1));
}

/* ------------------------------
 * - zip_extract_file_to_memory -
 * ------------------------------ */

static int
zip_extract_file_to_memory (const char *archive, const char *file,
			    char *buf, const unsigned long buflen)
{
  if (!zip_open(archive))
    return(0);

  /* Find the requested file - 0 = default case sensitivity */
  if (unzLocateFile(zipfile, file, 0) != UNZ_OK)
    return(0);

  /* Now read it into 'buf' */
  if (unzOpenCurrentFile(zipfile) != UNZ_OK)
    return(0);

  /* TODO: Can I assume that this will read all data? */
  if (unzReadCurrentFile(zipfile, buf, buflen) != buflen)
    return(0);

  unzCloseCurrentFile(zipfile);

  return(1);
}
